/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   fma_locks.c
 *    FreeM Administrative Tool - Locks
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2020 Coherent Logic Development LLC
 *
 *   Last modified: 13 Sep 2020
 * 
 **/

#if !defined(_SCO_DS) && !defined(__AMIGA) && !defined(_AIX)
# include <getopt.h>
#endif

#include "fmadm.h"

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>


int unlock (char *locktab, long pid);
long int tell (int fd);

int fma_locks_list (int optc, char **opts)
{
    short fd;           /* file descriptor */
    char line[STRLEN];  /* line of file */
    char c;             /* a character from the file */
    int i = 0;          /* counter */
    int l_pid;          /* pid owning lock */
    int l_typ;          /* lock type (LOCK or ZALLOCATE) */
    char t_nam[STRLEN]; /* buffer for building name */
    char l_nam[STRLEN]; /* buffer for lock name */
    int ct = 0;         /* number of locks found */
    int slots = 0;      /* slot count */

    if ((fd = open (fma_locktab, 0)) == -1) {
        fprintf (stderr, "fmadm:  could not open LOCK table %s\n", fma_locktab);
        return 1;
    }

    printf ("\nFreeM Lock Table\n");
    printf ("----------------\n\n");
    printf ("Namespace:     %s\n", fma_namespace);
    printf ("Lock Table:    %s\n\n", fma_locktab);


    printf ("%-10s %-10s %-10s\n", "PID", "TYPE", "LOCK NAME");
    printf ("%-10s %-10s %-10s\n", "---", "----", "---------");

    lseek (fd, 0L, 0);
    
    while (1) {
        
        read (fd, line, 3);

        l_pid = UNSIGN (line[0]) * 256 + UNSIGN (line[1]);
        l_typ = line[2];

        i = 0;
        do {

            if ((read (fd, &c, 1)) == 0) goto done;

            if (isprint (c)) t_nam[i++] = c;

        } while (c != EOL);

        zname (l_nam, t_nam);

        if (l_pid != 0) {
            printf ("%-10d %-10s %-10s\n", l_pid, l_typ == 'D' ? "ZALLOCATE" : "LOCK", l_nam);
            ct++;
        }

        slots++;

    }

done:
    printf ("\n    - %d entries found\n    - %d slot(s) in table\n\n", ct, slots);

    return 0;
} /* fma_locks_list() */

int fma_locks_remove (int optc, char **opts)
{

    int ct;
    int c;
    int option_index = 0;
    long pid = 0;

#if !defined(_SCO_DS) && !defined(__AMIGA) && !defined(_AIX)

    struct option long_options[] = {        
        {"pid", required_argument, 0, 'p'},
        {0, 0, 0, 0}
    };

#endif

#if !defined(_SCO_DS) && !defined(__AMIGA) && !defined(_AIX)

    while(1) {
        
        c = getopt_long (optc, opts, "p:", long_options, &option_index);

        if (c == -1) break;

        switch (c) {

            case 'p':
                pid = atol (optarg);
                break;

        }
    }




#else

#endif

    if (pid == 0) {
        fprintf (stderr, "fmadm:  must supply -p or --pid argument\n");
        return 1;
    }

    printf ("\nRemoving all LOCKs for PID %d:\t", pid);

    ct = unlock (fma_locktab, pid);

    if (ct >= 0) {
        printf ("[%d MATCHES]\n\n", ct);
    }
    else {
        printf ("[ERROR]\n\n");
    }

    return ct >= 0 ? 0 : 1;

} /* fma_locks_remove() */

/* unLOCK all entries of pid */
int unlock (char *locktab, long pid)           
{

    short ltab;         /* file descriptor for locktab */
    int cpid;
    char entry[256];
    long int r_pos;     /* position to read  */
    long int w_pos;     /* position to write */
    int i;
    int j;

    /* open locktab, quit if nothing to be done */
    if ((ltab = open (locktab, 2)) == -1) {
        fprintf (stderr, "fmadm:  could not open LOCK table %s\n", fma_locktab);
        return -1;
    }
    /* request exclusive access to locktab (read & write) */
    
    locking (ltab, 1, 0L);

    /*      free all of your own locks; we do it by copying the whole stuff */
    /*      within 'locktab' omitting the old entries under our PID         */
    
    j = 0;              /* count your old entries */
    
    lseek (ltab, 0L, 0);
    w_pos = 0L;

    for (;;) {

        read (ltab, entry, 2);
        
        cpid = UNSIGN (entry[0]) * 256 + UNSIGN (entry[1]);
        
        if (cpid == 0) {
            
            lseek (ltab, w_pos, 0);
            write (ltab, entry, 2);
            
            break;

        }

        i = 1;
        do {

            read (ltab, &entry[++i], 1);

        } while (entry[i] != EOL);

        i++;
        
        if (cpid != pid) {
            
            if (j) {

                r_pos = tell (ltab);

                lseek (ltab, w_pos, 0);
                write (ltab, entry, (unsigned) i);
                lseek (ltab, r_pos, 0);

            }

            w_pos += i;

        } 
        else {
            j++;
        }

    }

    locking (ltab, 0, 0L);
    close (ltab);

    return j;

} /* lock() */